= sycl_ext_oneapi_invoke_simd
:source-highlighter: coderay
:coderay-linenums-mode: table

// This section needs to be after the document title.
:doctype: book
:toc2:
:toc: left
:encoding: utf-8
:lang: en

:blank: pass:[ +]

// Set the default source code type in this document to C++,
// for syntax highlighting purposes.  This is needed because
// docbook uses c++ and html5 uses cpp.
:language: {basebackend@docbook:c++:cpp}

== Introduction

NOTE: Khronos(R) is a registered trademark and SYCL(TM) and SPIR(TM) are trademarks of The Khronos Group Inc.  OpenCL(TM) is a trademark of Apple Inc. used by permission by Khronos.

This extension introduces a mechanism to invoke explicitly vectorized functions
written using SIMD types from a SYCL kernel.

== Notice

Copyright (c) 2021-2022 Intel Corporation.  All rights reserved.

== Status

Experimental

This is an experimental extension specification, intended to provide early
access to features and gather community feedback. Interfaces defined in this
specification are implemented in DPC++, but they are not finalized and may
change incompatibly in future versions of DPC++ without prior notice. Shipping
software products should not rely on APIs defined in this specification.

== Version

Revision: 4

== Contacts

John Pennycook, Intel (john 'dot' pennycook 'at' intel 'dot' com) +
Roland Schulz, Intel (roland 'dot' schulz 'at' intel 'dot' com) +
Jason Sewall, Intel (jason 'dot' sewall 'at' intel 'dot' com) +

== Dependencies

This extension is written against the SYCL 2020 specification, Revision 4 and the following extensions:

- SPV_INTEL_function_pointers
- sycl_ext_oneapi_uniform

== Feature Test Macro

This extension provides a feature-test macro as described in the core SYCL
specification section 6.3.3 "Feature test macros".  Therefore, an
implementation supporting this extension must predefine the macro
`SYCL_EXT_ONEAPI_INVOKE_SIMD` to one of the values defined in the table below.
Applications can test for the existence of this macro to determine if the
implementation supports this feature, or applications can test the macro's
value to determine which of the extension's APIs the implementation supports.

[%header,cols="1,5"]
|===
|Value |Description
|1     |Initial extension version.  Base features are supported.
|===

== Overview

Standard C++ functions are treated as SPMD functions in SYCL. "Horizontal" SIMD operations (e.g.
shuffles) can be expressed as group functions. If the user wants to tune for
the SIMD instruction set of a particular device, then they are dependent on an
implementation providing a sub-group extension that exposes the instructions
that they would like to use. The user might want to express their function
using SIMD types because they need to use intrinsics or inline ASM, have legacy
code that uses SIMD types, or are more familiar with that programming model.
This extension adds support for an execution model in which SIMD parallelism is
expressed using SIMD types, with clear delineation between SPMD and SIMD code regions through an invocation interface also introduced by the extension.

=== Defining a SIMD Function

Uniform arguments are represented by scalars, and varying arguments by explicit
vector types.  The explicit vector types are based on the SIMD classes
from [Parallelism TS2](http://wg21.link/N4808), and support all of the functions
operating on SIMD types as defined there. Additionally, each of the functions
defined in Sections 4.17.5, 4.17.6 and 4.17.7 of the SYCL 2020 specification
are overloaded to accept SIMD classes; these overloads operate element-wise on
the contents of the SIMD class.

[source, c++]
----
namespace sycl {
namespace ext {
namespace oneapi {
namespace experimental {

namespace simd_abi {
template <class T, int N>
using native_fixed_size = /* implementation-defined */;
} // namespace simd_abi

template <class T, int N>
using simd = std::experimental::simd<T, simd_abi::native_fixed_size<N>>;

template <class T, int N>
using simd_mask = std::experimental::simd_mask<T, simd_abi::native_fixed_size<N>>;

} // namespace experimental
} // namespace oneapi
} // namespace ext
} // namespace sycl
----

We introduce a new `native_fixed_size` ABI which combines elements of the
existing `native` and `fixed_size` ABIs defined in Parallelism TS2;
implementations are expected to use the most efficient representation
of a `simd<T, N>` type (e.g. using multiple registers rather than an array).
We expect implementations of this ABI to build on Clang's `ext_vector_type`,
rather than specifying the register types directly.

Each of the following must hold true for a SIMD function and its invoke_simd
call site:
- the value of `N` of each `simd` or `simd_mask` formal parameter corresponding to a non-uniform
SPMD actual argument (if any) must be the same
- if the return type of the SIMD function is `simd<T, M>` or `simd_mask<T, M>`,
  then `M` must match `N` above
Otherwise, the SIMD function cannot be invoked from an SPMD kernel and the
implementation must issue a diagnostic in this case.
The `N` value represents the sub-group size of the calling kernel.

The example below shows a simple SIMD function that scales all elements of
an 8-wide SIMD type `x` by a scalar value `n`.

[source, c++]
----
sycl::ext::oneapi::experimental::simd<float, 8> scale(sycl::ext::oneapi::experimental::simd<float, 8> x, float n)
{
  return x * n;
}
----

=== Invoking a SIMD Function

The `invoke_simd` free function invokes a SIMD function using all work-items
in a `sub_group`.  The `invoke_simd` interface marshals data between the
SPMD context of the calling kernel and the SIMD context of the callee,
converting arguments and return values between scalar and SIMD types as
appropriate.

[source, c++]
----
namespace sycl {
namespace ext {
namespace oneapi {
namespace experimental {

template <typename Callable, typename...T>
auto invoke_simd(sub_group sg, Callable&& f, T... args);

} // namespace experimental
} // namespace oneapi
} // namespace ext
} // namespace sycl
----

The `Callable` passed to `invoke_simd` must be a C++ callable that can be
invoked with the same number of arguments specified in the `args` parameter
pack. `Callable` may be a function object, a lambda,
or a function pointer (if the device supports `SPV_INTEL_function_pointers`).
`Callable` must be an immutable callable with the same type and state for all
work-items in the sub-group, otherwise behavior is undefined.

NOTE: When using a lambda `Callable`, it is very easy for a user to
unintentionally capture non-uniform variables and violate the requirements
from the previous paragraph. This form of `invoke_simd` should be used with
caution.

Each formal parameter of the invoked `Callable` must be a scalar type, a SIMD type,
or a `std::tuple` of these types. The length of each SIMD type must match the
sub-group size of the kernel; calling a SIMD function with mismatched lengths
is an error, and an implementation must issue a diagnostic.

To enable overloading in cases where the `Callable` accepts either a scalar
type or a SIMD type, we leverage the `sycl::ext::oneapi::experimental::uniform` wrapper
type from the `sycl_ext_oneapi_uniform` extension.

Each argument in the `args` parameter pack must be an arithmetic type, a
trivially copyable type wrapped in a `sycl::ext::oneapi::experimental::uniform`, or a
`std::tuple` of these types.  Arguments may not be pointers or references,
but pointers (like any other non-arithmetic type) may be passed if wrapped in a
`sycl::ext::oneapi::experimental::uniform`.  Any such pointer value must point to memory
that is accessible by all work-items in the sub-group (i.e. the pointer must
point to an allocation in local or global memory).  The address space for such
pointers can be local, global or generic; if a generic pointer that points
to an allocation in private memory is passed as an argument, the behavior is
undefined.

NOTE: The restrictions above may prevent SYCL kernels from directly invoking
pre-existing SIMD functions (e.g. because they expect a vector of pointers).
The user is required to define a wrapper function to expose a compatible
interface in such cases (e.g. a wrapper function that converts a uniform base
pointer and offsets into a vector of pointers).

In order to invoke the SIMD function, `invoke_simd` converts each argument
in the `args` parameter pack according to the following rules:

- Arguments of type `bool` are converted to type
  `sycl::ext::oneapi::experimental::simd_mask<T, N>`, where `N` is the sub-group size of
  the calling kernel, and `T` is a user-specified type supported by
  `sycl::ext::oneapi::experimental::simd_mask`.
  Element `i` of the SIMD type represents the value from
  the work-item with sub-group local ID `i`.

- All other arithmetic arguments of type `T` are converted to type
  `sycl::ext::oneapi::experimental::simd<T, N>`, where `N` is the sub-group size of the
  calling kernel.  Element `i` of the SIMD type represents the value from the
  work-item with sub-group local ID `i`.

- Arguments of type `sycl::ext::oneapi::experimental::uniform<T>` are converted to type `T`.
  Conversion follows the same rules as the implicit conversion `operator T()`
  from the `sycl::ext::oneapi::experimental::uniform<T>` class; if the return value of
  `operator T()` would be undefined, the value of the scalar variable passed
  to the SIMD function is undefined.

- Arguments of type `std::tuple` are mapped as if each member of the tuple
  was mapped individually (e.g. a `std::tuple<float, int>` becomes a
  `std::tuple<sycl::ext::oneapi::experimental::simd<float, N>, sycl::ext::oneapi::experimental::simd<int, N>>`.

After this mapping has taken place, an appropriate variant of `Callable` is
selected based on standard overload resolution.

NOTE: If overload behavior is desired, a user must pass an overload set to
`invoke_simd` (e.g. a `Callable` with multiple call operators).  When the name
of a function is passed to `invoke_simd` it decays to a function pointer; if
the name of an overloaded function is passed to `invoke_simd`, the compiler
cannot deduce the type of `Callable`.

The return value of an `invoke_simd` function is subject to the same mapping
rules in reverse:

- Return values of type `sycl::ext::oneapi::experimental::simd_mask<T, N>` are converted
  to `bool`, and the value in element `i` of the SIMD type is returned to the
  work-item with sub-group local ID `i`.

- Return values of all other arithmetic types `sycl::ext::oneapi::experimental::simd<T, N>`
  are converted to `T`, and the value in element `i` of the SIMD type is
  returned to the work-item with sub-group local ID `i`.

- Return values of type `sycl::ext::oneapi::experimental::uniform<T>` are not anyhow converted,
  and broadcast to each work-item; every work-item in the sub-group receives
  the same value.
  NOTE: `sycl::ext::oneapi::experimental::uniform<T>` return type is the way to return
  a uniform value of `simd` or `simd_mask` type.

- Return values of type `T` are converted to `sycl::ext::oneapi::experimental::uniform<T>`,
  and broadcast to each work-item; every work-item in the sub-group receives
  the same value.

- Return values of type `std::tuple` are mapped as if each member of the tuple
  was mapped individually (e.g. a `std::tuple<sycl::ext::oneapi::experimental::simd<float, N>, sycl::ext::oneapi::experimental::simd<int, N>>`
  becomes a `std::tuple<float, int>`).

Note that these conversion rules do not permit special types (e.g. `sycl::nd_item`)
being passed between SPMD and SIMD contexts, because they cannot be declared uniform.
If an application bypasses this restriction (e.g. by wrapping an unsupported
type, or passing an unsupported type via a pointer) then behavior is undefined.
This restriction prevents a SIMD function from calling functions that are only
well-defined in SPMD contexts (e.g. sub-group barriers).  If a SIMD function
requires access to a member variable of a SYCL class, the value of the
variable should be passed via a dedicated argument (e.g. the value returned
by `sub_group::get_local_id()[0]` could be passed as an integer to a
`Callable` expecting a `sycl::ext::oneapi::experimental::simd<uint32_t, N>`).

NOTE: An implementation must be able to determine the sub-group size of the SIMD
function object to perform argument type conversion.
The determined subgroup size must be equal to the value of the caller's `sycl::reqd_sub_group_size`
attribute (if present) and consistent with the argument conversion rules defined
by this specification. Otherwise, the implementation must issue a proper diagnostic.

The following algorithm to determine the sub-group size may be employed:
- If any of the actual SPMD arguments is non-uniform, the specification requires
that the corresponding SIMD argument must have `simd<T, N>` or `simd_mask<T, N>`
type, and `N` must be equal to the the sub-group size. The implementation may
iterate through the supported sub-group sizes at compile time and check whether the
function object is invocable with given sub-group size. There must be
exactly one such sub-group size and it must match `N` or the user program is not
well-formed.
- Otherwise, if the return type of the SIMD target function is `simd<T, N>` or
`simd_mask<T, N>`, then the sub-group size is `N`.
- Otherwise (all arguments and the return type are uniform) the SIMD target is
fully uniform and it can be invoked with any sub-group size.

The `invoke_simd` function has the same requirements as other group functions
(as defined in Section 4.17.3 of the SYCL 2020 specification). A call to
`invoke_simd` must be encountered in converged control flow by all work-items
in the group, and the call acts as a synchronization point -- the `Callable` is
not invoked until all work-items reach the call to `invoke_simd`, and all
work-items must wait for the `Callable` to complete before continuing
execution.

The example below shows how to invoke a simple SIMD function that scales all
elements of a SIMD type `x` by a scalar value `n`, both with and without a
SIMD mask parameter.

[source, c++]
----
sycl::ext::oneapi::experimental::simd<float, 8> scale(sycl::ext::oneapi::experimental::simd<float, 8> x, float n)
{
  return x * n;
}

sycl::ext::oneapi::experimental::simd<float, 8> masked_scale(sycl::ext::oneapi::experimental::simd<float, 8> x, float n, sycl::ext::oneapi::experimental::simd_mask<float, 8> mask)
{
  std::experimental::where(mask, x) *= n;
  return x;
}

q.parallel_for(..., sycl::nd_item<1> it) [[sycl::reqd_sub_group_size(8)]]
{
  sycl::sub_group sg = it.get_sub_group();
  float x = ...;
  float n = ...;

  // invoke SIMD function
  // x values from each work-item are combined into a simd<float, 8>
  float y = sycl::ext::oneapi::experimental::invoke_simd(sg, scale, x, sycl::ext::oneapi::experimental::uniform(n));

  // invoke SIMD function with a mask parameter
  // x values from each work-item be a simd<float, 8>
  // mask values from each work-item are combined into a simd_mask<float, 8>
  bool mask = (it.get_local_id(0) % 2);
  float z = sycl::ext::oneapi::experimental::invoke_simd(sg, masked_scale, x, sycl::ext::oneapi::experimental::uniform(n), mask);
});
----

The examples below demonstrates how invoking a SIMD function behaves in the
presence of function overloads.

[source, c++]
----
sycl::ext::oneapi::experimental::simd<float, 8> scale(sycl::ext::oneapi::experimental::simd<float, 8> x, float n)
{
  return x * n;
}

sycl::ext::oneapi::experimental::simd<float, 8> scale(sycl::ext::oneapi::experimental::simd<float, 8> x, sycl::ext::oneapi::experimental::simd<float, 8> n)
{
  return x * n;
}

q.parallel_for(..., sycl::nd_item<1> it) [[sycl::reqd_sub_group_size(8)]]
{
  sycl::sub_group sg = it.get_sub_group();
  float x = ...;
  float n = ...;

  // invoke SIMD function
  // which overload to call must be expressed explicitly, here we use invoke_simd's template argument
  float y = sycl::ext::oneapi::experimental::invoke_simd<sycl::ext::oneapi::experimental::simd<float, 8>(sycl::ext::oneapi::experimental::simd<float, 8>, float)>(sg, scale, x, sycl::ext::oneapi::experimental::uniform(n));
});
----

[source, c++]
----
struct scale
{
  sycl::ext::oneapi::experimental::simd<float, 8> operator()(sycl::ext::oneapi::experimental::simd<float, 8> x, float n)
  {
    return x * n;
  }

  sycl::ext::oneapi::experimental::simd<float, 8> operator()(sycl::ext::oneapi::experimental::simd<float, 8> x, sycl::ext::oneapi::experimental::simd<float, 8> n)
  {
    return x * n;
  }
};

q.parallel_for(..., sycl::nd_item<1> it) [[sycl::reqd_sub_group_size(8)]]
{
  sycl::sub_group sg = it.get_sub_group();
  float x = ...;
  float n = ...;

  // invoke SIMD function
  // the correct overload of scale's call operator will be selected after mapping to SIMD types
  // note that invoke_simd accepts an instance of scale
  float y = sycl::ext::oneapi::experimental::invoke_simd(sg, scale(), x, sycl::ext::oneapi::experimental::uniform(n));
});
----

==== Execution Model

Execution of a SIMD function must produce the same result as-if it had been
executed by a single work-item executing SIMD instructions.

The following are all examples of valid implementations of this model:

- Mapping SIMD operations directly to SIMD instructions.

- Mapping logical SIMD operations wider than the native SIMD width to multiple
  SIMD instructions.

- Mapping each element of a SIMD type to an individual work-item in a
  sub-group, with appropriate barriers and fences generated between SIMD
  operations.

- Mapping SIMD operations to an unrolled loop executed by a single work-item.

Since this execution model guarantees SIMD-like behavior, there is no need for
the user to insert any form of explicit synchronization functions to ensure
memory consistency across SIMD lanes.

Due to potential differences in execution model, a SIMD function cannot call
any function that was compiled for a SPMD context. `SYCL_EXTERNAL` functions
defined in a different translation unit to the SIMD function are not required
to be compiled for SIMD contexts by default, and as a result such functions
are incompatible with SIMD functions by default.

NOTE: Implementations are encouraged to throw an error in cases where this
behavior can be detected, and to do so as early as possible.

This extension does not define a mechanism for requesting a `SYCL_EXTERNAL`
function to be compiled for a SIMD context. Any such mechanism and its
interaction with `invoke_simd` is implementation-defined.

NOTE: An explicit mechanism is required because the presence or absence of a
`sycl::ext::oneapi::experimental::simd` class is not sufficient to determine
the execution model associated with a function. For example, the execution
model for a function accepting only scalar parameters and performing only
scalar arithmetic is ambiguous: it could be called by each work-item in SPMD
code, or called as-if by a single work-item in a SIMD function.

== Issues

. Should we allow reference arguments?
+
--
*RESOLVED*: No.  Unlike pointers, it is less clear that references carry
address space information, and users may not expect a `T&` argument to behave
like a pointer to the generic address space.  Users are also more likely to try
and pass a reference to a variable in the private address space than a reference
to a variable in the local or global address spaces, which would lead to
undefined behavior and issues that are difficult to debug.
--

. How should returning multiple values and passing user-defined structs be handled?
+
--
*UNRESOLVED*: `std::tuple` addresses both issues from a user interface perspective,
but this aspect of the design may need to be revisited as we gain implementation experience.
--

. Should arguments require trivially copyable or device copyable?
+
--
*UNRESOLVED*: Device copyable permits bitwise copies of a wider range of classes, but the
SYCL 2020 specification states that this only applies to inter-device transfers.
--

//. asd
//+
//--
//*RESOLUTION*: Not resolved.
//--
